@@ -1,15 +1,12 @@
-import os
-import subprocess as sp
-
 import dill
 import numpy as np
 import pandas as pd
-from sklearn.cluster import KMeans, DBSCAN
+from sklearn.cluster import KMeans
 from sklearn.decomposition import PCA
 from zedsuite.zutil import Llk, Lsmash, Prun, DrawPFSA
 from zedsuite.genesess import GenESeSS
 from zedsuite.quantizer import Quantizer
-from patternly._utils import RANDOM_NAME, UnionFind, DirectedGraph
+from patternly._utils import RANDOM_NAME, DirectedGraph
 
 
 class AnomalyDetection:
@@ -23,7 +20,7 @@ class AnomalyDetection:
         reduce_clusters=True,
         clustering_alg=None,
         quantize=True,
-        quantize_type="complex",
+        quantize_type='complex',
         eps=0.1,
         verbose=False
     ) -> None:
@@ -55,8 +52,8 @@ class AnomalyDetection:
         self.quantize_type = quantize_type
         self.eps = eps
         self.verbose = verbose
-        self.temp_dir = "zed_temp"
-        self.PFSA_prop_titles =  ["%ANN_ERR", "%MRG_EPS", "%SYN_STR", "%SYM_FRQ", "%PITILDE", "%CONNX"]
+        self.temp_dir = 'zed_temp'
+        self.PFSA_prop_titles =  ['%ANN_ERR', '%MRG_EPS', '%SYN_STR', '%SYM_FRQ', '%PITILDE', '%CONNX']
 
         # values calculated in self.fit()
         self.fitted = False                # whether model has been fit
@@ -74,6 +71,7 @@ class AnomalyDetection:
         # can be accessed after calling self.predict()
         self.cluster_llks = None           # list of llk of each prediction for each cluster
         self.closest_match = None          # cluster label of closest match for each prediction
+        self.predicted_active_pfsas = {}   # dictionary of each pfsa active at each unit of time in the sequence
 
 
     def fit(self, X, y=None):
@@ -114,7 +112,7 @@ class AnomalyDetection:
         """
 
         if not self.fitted:
-            raise ValueError("Model has not been fit yet")
+            raise ValueError("Model has not been fit yet.")
 
         data = None
         num_predictions = 0
@@ -123,16 +121,27 @@ class AnomalyDetection:
             # occurs when model is loaded from file
             if self.quantized_data is None:
                 raise ValueError("Original data not found. Pass data to predict().")
-            data = self.quantized_data.drop(columns=["cluster"], axis=1)
+            data = self.quantized_data.drop(columns=['cluster'], axis=1, errors='ignore')
+            # data.to_csv('error.csv', sep=' ', index=False, header=False)
             num_predictions = self.quantized_data.shape[0]
         else:
             data = self.__quantize(X)
             num_predictions = 1 if type(X) is pd.Series else data.shape[0]
 
-        cluster_llks = np.empty(shape=(self.n_clusters, num_predictions), dtype=np.float32)
+        cluster_llks = np.zeros(shape=(self.n_clusters, num_predictions), dtype=np.float32)
+
+        def count_clusters(row):
+            clusters = np.bincount(row)
+            return clusters.size
+        data['alphabet_size'] = data.apply(count_clusters, axis=1)
+
         for i in range(self.n_clusters):
-            # cluster_llks[i] = np.asarray(Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run(), dtype=np.float32)
-            llks = Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run()
+            curr_alphabet_size = len(self.cluster_PFSA_info[i]['%SYM_FRQ'])
+            curr_data = data[data['alphabet_size'] <= curr_alphabet_size].drop(columns=['alphabet_size'], axis=1)
+            llks = Llk(data=curr_data, pfsafile=self.cluster_PFSA_files[i]).run()
+            llks = pd.concat(
+                [data[0], pd.DataFrame(llks, index=curr_data.index, columns=['llks'])], axis=1
+            ).fillna(float('inf'))['llks'].values.tolist()
             cluster_llks[i] = np.asarray(llks, dtype=np.float32)
 
         # consider to be anomaly if all llks above specified upper bound (X standard deviations above the mean)
@@ -141,14 +150,20 @@ class AnomalyDetection:
 
         self.cluster_llks = cluster_llks
         self.closest_match = np.argmin(cluster_llks, axis=0)
+        self.predicted_active_pfsas = {}
+        for i, pfsa in enumerate(self.closest_match):
+            if pfsa in self.predicted_active_pfsas:
+                self.predicted_active_pfsas[pfsa].append(i)
+            else:
+                self.predicted_active_pfsas[pfsa] = [i]
 
         if len(predictions) == 1:
-            predictions = predictions[0]
+            return predictions[0]
 
         return predictions
 
 
-    def save_model(self, path="patternly_model.dill"):
+    def save_model(self, path='patternly_model.dill'):
         """ Save model to file
 
         Args:
@@ -156,37 +171,37 @@ class AnomalyDetection:
         """
 
         if not self.fitted:
-            raise ValueError("Model has not been fit yet")
+            raise ValueError("Model has not been fit yet.")
 
         metadata = {
-            "modeltype": type(self),
-            "user_params": {
-                "anomaly_sensitivity": self.anomaly_sensitivity,
-                "n_clusters": self.n_clusters,
-                "reduce_clusters": self.reduce_clusters,
-                "clustering_alg": self.clustering_alg,
-                "quantize": self.quantize,
-                "quantize_type": self.quantize_type,
-                "eps": self.eps,
-                "verbose": self.verbose,
+            'modeltype': type(self),
+            'user_params': {
+                'anomaly_sensitivity': self.anomaly_sensitivity,
+                'n_clusters': self.n_clusters,
+                'reduce_clusters': self.reduce_clusters,
+                'clustering_alg': self.clustering_alg,
+                'quantize': self.quantize,
+                'quantize_type': self.quantize_type,
+                'eps': self.eps,
+                'verbose': self.verbose,
             },
-            "fitted_params": {
-                "quantizer_parameters": None if self.quantizer is None else self.quantizer.parameters ,
-                "quantizer_feature_order": None if self.quantizer is None else self.quantizer._feature_order,
-                "cluster_labels": self.cluster_labels,
-                "cluster_counts": self.cluster_counts,
-                "cluster_PFSA_info": self.cluster_PFSA_info,
-                "PFSA_llk_means": self.PFSA_llk_means.tolist(),
-                "PFSA_llk_stds": self.PFSA_llk_stds.tolist()
+            'fitted_params': {
+                'quantizer_parameters': None if self.quantizer is None else self.quantizer.parameters ,
+                'quantizer_feature_order': None if self.quantizer is None else self.quantizer._feature_order,
+                'cluster_labels': self.cluster_labels,
+                'cluster_counts': self.cluster_counts,
+                'cluster_PFSA_info': self.cluster_PFSA_info,
+                'PFSA_llk_means': self.PFSA_llk_means.tolist(),
+                'PFSA_llk_stds': self.PFSA_llk_stds.tolist()
             }
         }
 
-        with open(path, "wb") as f:
+        with open(path, 'wb') as f:
             dill.dump(metadata, f)
 
 
     @staticmethod
-    def load_model(path="patternly_model.dill"):
+    def load_model(path='patternly_model.dill'):
         """ Load saved model
 
             Args:
@@ -196,24 +211,24 @@ class AnomalyDetection:
                 AnomalyDetection or StreamingDetection: loaded model
         """
 
-        with open(path, "rb") as f:
+        with open(path, 'rb') as f:
             metadata = dill.load(f)
 
         model = None
-        if metadata["modeltype"] is AnomalyDetection:
-            model = AnomalyDetection(**metadata["user_params"])
+        if metadata['modeltype'] is AnomalyDetection:
+            model = AnomalyDetection(**metadata['user_params'])
         else:
-            model = StreamingDetection(**metadata["user_params"])
+            model = StreamingDetection(**metadata['user_params'])
 
-        if model.quantize and model.quantize_type == "complex":
+        if model.quantize and model.quantize_type == 'complex':
             model.quantizer = Quantizer(n_quantizations=1, eps=-1)
-            model.quantizer.parameters = metadata["fitted_params"]["quantizer_parameters"]
-            model.quantizer._feature_order = metadata["fitted_params"]["quantizer_feature_order"]
-        model.cluster_labels = metadata["fitted_params"]["cluster_labels"]
-        model.cluster_counts = metadata["fitted_params"]["cluster_counts"]
-        model.cluster_PFSA_info = metadata["fitted_params"]["cluster_PFSA_info"]
-        model.PFSA_llk_means = np.asarray(metadata["fitted_params"]["PFSA_llk_means"])
-        model.PFSA_llk_stds = np.asarray(metadata["fitted_params"]["PFSA_llk_stds"])
+            model.quantizer.parameters = metadata['fitted_params']['quantizer_parameters']
+            model.quantizer._feature_order = metadata['fitted_params']['quantizer_feature_order']
+        model.cluster_labels = metadata['fitted_params']['cluster_labels']
+        model.cluster_counts = metadata['fitted_params']['cluster_counts']
+        model.cluster_PFSA_info = metadata['fitted_params']['cluster_PFSA_info']
+        model.PFSA_llk_means = np.asarray(metadata['fitted_params']['PFSA_llk_means'])
+        model.PFSA_llk_stds = np.asarray(metadata['fitted_params']['PFSA_llk_stds'])
         # print(model.PFSA_llk_means)
         # print(model.PFSA_llk_stds)
 
@@ -221,8 +236,8 @@ class AnomalyDetection:
         model.cluster_PFSA_files = []
         for i in range(model.n_clusters):
             model.cluster_PFSA_files.append(RANDOM_NAME(path=model.temp_dir))
-            with open(model.cluster_PFSA_files[i], "w") as f:
-                f.write(f"{model.__format_PFSA_info(model.cluster_PFSA_info[i])}")
+            with open(model.cluster_PFSA_files[i], 'w') as f:
+                f.write(f'{model.__format_PFSA_info(model.cluster_PFSA_info[i])}')
                 # print(model.__format_PFSA_info(model.cluster_PFSA_info[i]))
         # model.generate_PFSA_pngs()
 
@@ -235,7 +250,7 @@ class AnomalyDetection:
         """ Print PFSAs found for each cluster """
 
         if not self.fitted:
-            raise ValueError("Model has not been fit yet")
+            raise ValueError("Model has not been fit yet.")
 
         for i in range(self.n_clusters):
             print(f"Cluster {i} PFSA:")
@@ -274,15 +289,15 @@ class AnomalyDetection:
         if type(X) is pd.Series:
             X = X.copy().to_frame().reset_index(drop=True).T
 
-        if self.quantize_type == "simple":
+        if self.quantize_type == 'simple':
             # basic differentiation
             X = X.astype(float).diff(axis=1).fillna(0)
             return X.apply(lambda row : row.apply(lambda n : 1 if n > 0 else 0), axis=1)
-        elif self.quantize_type == "simple-second":
+        elif self.quantize_type == 'simple-second':
             # second derivative
             X = X.astype(float).diff(axis=1).fillna(0).diff(axis=1).fillna(0)
             return X.apply(lambda row : row.apply(lambda n : 1 if n > 0 else 0), axis=1)
-        elif self.quantize_type == "complex":
+        elif self.quantize_type == 'complex':
             # use cythonized quantizer binary
             if not self.fitted:
                 self.quantizer = Quantizer(n_quantizations=1, epsilon=-1)
@@ -306,7 +321,7 @@ class AnomalyDetection:
             self.dist_matrix = self.quantized_data
         else:
             self.dist_matrix = pd.DataFrame(
-                Lsmash(data=self.quantized_data, data_type="symbolic", sae=False).run(),
+                Lsmash(data=self.quantized_data, data_type='symbolic', sae=False).run(),
                 dtype=np.float32
             )
 
@@ -335,10 +350,23 @@ class AnomalyDetection:
         cluster_labels = [cluster_rank[cluster] for cluster in cluster_labels]
         cluster_counts = [cluster_counts[cluster] for cluster in np.argsort(cluster_counts)][::-1]
 
+        # def adjust_clusters(row):
+        #     bins = np.bincount(row)
+        #     offset = [0 for _ in range(len(bins))]
+        #     preceeding_zeros = 0
+        #     for i, bin in enumerate(bins):
+        #         if bin == 0:
+        #             preceeding_zeros += 1
+        #         offset[i] = preceeding_zeros
+        #     return row.apply(lambda val: val - offset[val])
+        # self.quantized_data = self.quantized_data.apply(adjust_clusters, axis=1)
+
         self.n_clusters = n_clusters
         self.cluster_counts = cluster_counts
         self.cluster_labels = cluster_labels
-        self.quantized_data["cluster"] = cluster_labels
+        self.quantized_data['cluster'] = cluster_labels
+        # self.quantized_data['alphabet_size'] = self.quantized_data.apply(count_clusters, axis=1)
+        # self.quantized_data[self.quantized_data['cluster'] == 0].to_csv('firstcluster.csv', sep=' ', header=False, index=False)
 
 
     def __calculate_cluster_PFSAs(self):
@@ -349,7 +377,9 @@ class AnomalyDetection:
         for i in range(self.n_clusters):
             if self.verbose:
                 print(f"Generating cluster PFSA {i + 1}/{self.n_clusters}...")
-            cluster_data = self.quantized_data[self.quantized_data["cluster"] == i].drop(columns=["cluster"], axis=1)
+            cluster_data = self.quantized_data[self.quantized_data['cluster'] == i].drop(
+                columns=['cluster'], axis=1
+            )
             cluster_PFSA_file = RANDOM_NAME(path=self.temp_dir)
             dot_file = RANDOM_NAME(path=self.temp_dir)
             cluster_PFSA_files.append(cluster_PFSA_file)
@@ -357,11 +387,12 @@ class AnomalyDetection:
             alg = GenESeSS(
                 data=cluster_data,
                 outfile=cluster_PFSA_file,
-                data_type="symbolic",
-                data_dir="row",
+                data_type='symbolic',
+                data_dir='row',
                 force=True,
                 eps=self.eps,
                 dot=dot_file,
+                configfile='./patternly/config.cfg'
             )
             # TODO: add error handling for when genESeSS fails to find a PFSA
             PFSA_found = alg.run()
@@ -393,22 +424,23 @@ class AnomalyDetection:
         if self.verbose:
             print("Attempting to reduce clusters...")
 
-        # generate representative sequences for each cluster PFSA
-        #  seq_len = 1000
-        #  seqs_per_pfsa = 100
-        #  all_clusters_seqs = [
-            #  Prun(pfsafile=pfsa, data_len=seq_len, num_repeats=seqs_per_pfsa).run()
-            #  for pfsa in self.cluster_PFSA_files
-        #  ]
-
         all_cluster_likelihoods = np.empty(shape=(self.n_clusters, self.n_clusters), dtype=np.float32)
         all_ranked_likelihoods = np.empty(shape=(self.n_clusters, self.n_clusters), dtype=np.int32)
 
         for i in range(self.n_clusters):
             cluster_llks = []
-            for pfsafile in self.cluster_PFSA_files:
-                cluster_data = self.quantized_data[self.quantized_data["cluster"] == i].drop(columns=["cluster"], axis=1)
-                cluster_llks.append(np.asarray(Llk(data=cluster_data, pfsafile=pfsafile).run(), dtype=np.float32))
+            for j in range(self.n_clusters):
+                curr_alphabet_size = len(self.cluster_PFSA_info[j]['%SYM_FRQ'])
+                cluster_data = self.quantized_data[self.quantized_data['cluster'] == i]
+                curr_data = cluster_data
+                # curr_data = cluster_data[cluster_data['alphabet_size'] <= curr_alphabet_size].drop(
+                    # columns=['cluster', 'alphabet_size'], axis=1
+                # )
+                llks = [] if curr_data.shape[0] == 0 else Llk(data=curr_data, pfsafile=self.cluster_PFSA_files[j]).run()
+                # llks = pd.concat(
+                    # [cluster_data[0], pd.DataFrame(llks, index=curr_data.index, columns=['llks'])], axis=1
+                # ).fillna(float('inf'))['llks'].values.tolist()
+                cluster_llks.append(np.asarray(llks, dtype=np.float32))
 
             # which cluster PFSA each sequence most likely maps back to
             closest_matches = np.argmin(cluster_llks, axis=0)
@@ -416,12 +448,14 @@ class AnomalyDetection:
             cluster_likelihoods = np.count_nonzero(
                 (closest_matches.reshape(-1, 1) == np.arange(self.n_clusters).reshape(1, -1)),
                 axis=0
-            ) / self.quantized_data[self.quantized_data["cluster"] == i].shape[0]
+            ) 
+            cluster_likelihoods = cluster_likelihoods / cluster_likelihoods.sum()
             # list of cluster PFSAs sorted in descending order of likelihood
             ranked_likelihoods = np.argsort(cluster_likelihoods)[::-1]
 
             all_cluster_likelihoods[i] = cluster_likelihoods
             all_ranked_likelihoods[i] = ranked_likelihoods
+            # print(all_cluster_likelihoods)
 
         for i in range(self.n_clusters):
             best_match = all_ranked_likelihoods[i][0]
@@ -437,7 +471,7 @@ class AnomalyDetection:
             self.n_clusters = new_n_clusters
             #  reduced_clusters = set(graph.roots)
             #  self.cluster_labels = [graph.roots[cluster] for cluster in self.cluster_labels]
-            #  self.quantized_data["cluster"] = self.cluster_labels
+            #  self.quantized_data['cluster'] = self.cluster_labels
             #  self.cluster_PFSA_files = [self.cluster_PFSA_files[i] for i in reduced_clusters]
             #  self.cluster_PFSA_info = [self.cluster_PFSA_info[i] for i in reduced_clusters]
             #  self.cluster_PFSA_pngs = [self.cluster_PFSA_pngs[i] for i in reduced_clusters]
@@ -461,11 +495,22 @@ class AnomalyDetection:
         PFSA_llk_means = np.empty(shape=self.n_clusters)
         PFSA_llk_stds = np.empty(shape=self.n_clusters)
         for i in range(self.n_clusters):
-            cluster_data = self.quantized_data[self.quantized_data["cluster"] == i].drop(columns=["cluster"], axis=1)
+            cluster_data = self.quantized_data[self.quantized_data['cluster'] == i].drop(
+                columns=['cluster'], axis=1
+            )
             llks = np.asarray(Llk(data=cluster_data, pfsafile=self.cluster_PFSA_files[i]).run(), dtype=np.float32)
             PFSA_llk_means[i] = np.mean(llks)
             PFSA_llk_stds[i] = np.std(llks, ddof=1)
 
+        # for i in range(self.n_clusters):
+        #     curr_alphabet_size = len(self.cluster_PFSA_info[i]['%SYM_FRQ'])
+        #     curr_data = data[data['alphabet_size'] <= curr_alphabet_size].drop(columns=['cluster', 'alphabet_size'], axis=1)
+        #     llks = Llk(data=curr_data, pfsafile=self.cluster_PFSA_files[i]).run()
+        #     llks = pd.concat(
+        #         [data[0], pd.DataFrame(llks, index=curr_data.index, columns=['llks'])], axis=1
+        #     ).fillna(float('inf'))['llks'].values.tolist()
+        #     cluster_llks[i] = np.asarray(llks, dtype=np.float32)
+
         self.PFSA_llk_means, self.PFSA_llk_stds = PFSA_llk_means, PFSA_llk_stds
 
 
@@ -480,29 +525,29 @@ class AnomalyDetection:
                 str: formatted PFSA information
         """
 
-        indent = indent_level * " "
+        indent = indent_level * ' '
 
-        syn_str_vals = ""
-        if PFSA_info["%SYN_STR"] is not None:
-            for syn_str in PFSA_info["%SYN_STR"]:
-                syn_str_vals += (f"{syn_str} ")
+        syn_str_vals = ''
+        if PFSA_info['%SYN_STR'] is not None:
+            for syn_str in PFSA_info['%SYN_STR']:
+                syn_str_vals += (f'{syn_str} ')
 
-        sym_frq_vals = ""
-        for sym_frq in PFSA_info["%SYM_FRQ"]:
-            sym_frq_vals += (f"{sym_frq} ")
+        sym_frq_vals = ''
+        for sym_frq in PFSA_info['%SYM_FRQ']:
+            sym_frq_vals += (f'{sym_frq} ')
 
         pitilde_vals = indent
-        for row in PFSA_info["%PITILDE"]:
+        for row in PFSA_info['%PITILDE']:
             for i, val in enumerate(row):
-                suffix = " " if (i+1) < len(row) else f" \n{indent}"
-                pitilde_vals += (f"{val}{suffix}")
+                suffix = ' ' if (i+1) < len(row) else f' \n{indent}'
+                pitilde_vals += (f'{val}{suffix}')
         pitilde_vals = pitilde_vals[:-len(indent)] if indent_level > 0 else pitilde_vals
 
         connx_vals = indent
-        for row in PFSA_info["%CONNX"]:
+        for row in PFSA_info['%CONNX']:
             for i, val in enumerate(row):
-                suffix = " " if (i+1) < len(row) else f" \n{indent}"
-                connx_vals += (f"{val}{suffix}")
+                suffix = ' ' if (i+1) < len(row) else f' \n{indent}'
+                connx_vals += (f'{val}{suffix}')
         connx_vals = connx_vals[:-len(indent)] if indent_level > 0 else connx_vals
 
         return (
@@ -549,16 +594,16 @@ class StreamingDetection(AnomalyDetection):
             return super().predict(X_split_streams)
 
     
-    def save_model(self, path="patternly_model.dill"):
+    def save_model(self, path='patternly_model.dill'):
         super().save_model(path=path)
 
-        with open(path, "rb") as f:
+        with open(path, 'rb') as f:
             metadata = dill.load(f)
 
-        metadata["user_params"]["window_size"] = self.window_size
-        metadata["user_params"]["window_overlap"] = self.window_overlap
+        metadata['user_params']['window_size'] = self.window_size
+        metadata['user_params']['window_overlap'] = self.window_overlap
 
-        with open(path, "wb") as f:
+        with open(path, 'wb') as f:
             dill.dump(metadata, f)
 
 
@@ -571,147 +616,137 @@ class StreamingDetection(AnomalyDetection):
                 X (pd.Series): data to be split into streams
         """
 
+        if type(X) is not pd.DataFrame:
+            X = pd.DataFrame(X)
+
         def beg(i): return int((window_size * i) - (window_overlap * i))
         def end(i): return int(beg(i) + window_size)
         size = int(X.shape[0] // (window_size - window_overlap))
         return pd.concat(
             [X[beg(i):end(i)].reset_index(drop=True) for i in range(size)],
             axis=1
-        ).T.reset_index(drop=True)
+        ).T.reset_index(drop=True).dropna(how='any', axis=0)
 
 
-    # def visualize(self, num_plots, predictions):
-        # """ Visualize active clusters and anomalies in respect to the original data """
-        # pass
-        # TODO: abstract and generalize for common use cases
+class ContinuousStreamingDetection(StreamingDetection):
+    """ Tool for anomaly detection within a single continual data stream """
 
-        # import matplotlib.pyplot as plt
-        # from datetime import datetime
-        # plt.rcParams["figure.figsize"] = (20, num_plots * 8)
+    def __init__(self, **kwargs):
+        """
+        Args: see StreamingDetection
+        """
 
-        # test_start_ts = test_info["meas_date"].timestamp()
-        # for i, channel in enumerate(test_channels):
-            
-        #     plot_data = {}
-        #     plot_data[channel] = test_data
-        #     plot_data["ts"] = pd.DataFrame([datetime.fromtimestamp(test_start_ts + (j / hz)) for j in range(plot_data[channel].shape[0])])
-        #     df = pd.concat([plot_data[channel], plot_data["ts"]], axis=1)
-        #     df.columns = ["data", "ts"]
-            
-        #     # plt.subplot(len(channels), 1, i+1)
-        #     plt.subplot(NUM_CHANNELS, 1, i+1)
-        #     plt.plot(df["ts"], df["data"], color="green", markersize=2, zorder=1)
+        self.pattern_emergence_times = []
+        super().__init__(**kwargs)
 
-        #     # highlight true anomalies
-        #     beg = datetime.fromtimestamp(test_start_ts + seizure_start_time)
-        #     end = datetime.fromtimestamp(test_start_ts + seizure_end_time)
-        #     plt.axvspan(beg, end, color='black', alpha=0.6, lw=0, zorder=2)
 
-        #     # highlight anomalies
-        #     anoms = list(predictions[i][predictions[i][0] == True].index)
-        #     for index in anoms:
-        #         beg = plot_data["ts"].iloc[(index * self.window_size) - (index * self.window_overlap)]
-        #         end = plot_data["ts"].iloc[min(((index + 1) * self.window_size) - (index * self.window_overlap), plot_data["ts"].shape[0] - 1)]
-        #         plt.axvspan(beg, end, color='red', alpha=0.5, lw=0, zorder=3)
+    def fit_stream(self, X, *, clean=True):
+        """ Fit new stream of data
 
-        #     # highlight non-anomalies by cluster
-        #     colors = ["#7766ee", "blue", "darkgreen", "darkorange", 'cyan', "navy", "grey", "yellow", "white", "purple"]
-        #     # non_anomalies = predictions[predictions == False]
-        #     non_anomalies = list(predictions[i][predictions[i][0] == False].index)
-        #     for index in non_anomalies:
-        #         beg = plot_data["ts"].iloc[(index * self.window_size) - (index * self.window_overlap)]
-        #         end = plot_data["ts"].iloc[min(((index + 1) * self.window_size) - (index * self.window_overlap), plot_data["ts"].shape[0] - 1)]
-        #         color = colors[pipelines[i].closest_match[index]]
-        #         plt.axvspan(beg, end, color=color, alpha=0.25, lw=0, zorder=0)
+        Args:
+            X (pd.DataFrame or pd.Series): time series data stream to find anomalies
+            clean (bool, optional): whether to remove temp files (Default = True)
 
-        #     # nice formatting
-        #     ax = plt.gca()
-        #     plt.setp(ax.get_xticklabels(), fontsize=16, y=-.02)
-        #     plt.setp(ax.get_yticklabels(), fontsize=16)
-        #     ax.tick_params(axis='x', colors='.5')    
-        #     ax.tick_params(axis='y', colors='.5', pad=15)  
-        #     ax.yaxis.offsetText.set_fontsize(16)
-        #     ax.spines['bottom'].set_color('.75')
-        #     ax.spines['top'].set_color('.75') 
-        #     ax.spines['right'].set_color('.75')
-        #     ax.spines['left'].set_color('.75')
-        #     ax.set_title(f"{channel}", fontsize=16, y=1.02, color='.5')
+        Returns:
+            ContinuousStreamingDetection: fitted model
+        """
 
-        # plt.tight_layout(pad=2)
-        # plt.show()
+        if X is None:
+            raise ValueError("Please pass data stream to fit.")
 
-# class AnomalyDetectionBin(AnomalyDetection):
-#     def __init__(self, **kwargs):
-#         super().__init__(**kwargs)
+        self.pattern_emergence_times = []
+        X_split_streams = self.split_streams(X, self.window_size, self.window_overlap)
+        data = self._AnomalyDetection__quantize(X_split_streams)
+        num_predictions = data.shape[0]
 
-#     @staticmethod
-#     def llk(data, pfsa_file, *, clean=True, data_file=None):
-#         data_file = RANDOM_NAME(clean) if data_file is None else data_file
-#         data.to_csv(data_file, sep=" ", index=False, header=False)
-#         llk_binary = os.path.abspath("./bin/llk")
-#         llpos = llk_binary + " -s " + data_file + " -f " + pfsa_file
+        def count_clusters(row):
+            clusters = np.bincount(row)
+            return clusters.size
+        data['alphabet_size'] = data.apply(count_clusters, axis=1)
 
-#         try:
-#             df_toreturn = pd.DataFrame(
-#                 sp.check_output(llpos, shell=True, stderr=sp.DEVNULL).split()
-#             ).astype(float)
+        # if not fitted with previous data, create a PFSA from first stream
+        if not self.fitted:
+            self.n_clusters = 0
+            self.fitted = True
+            self.__add_to_PFSA_library(data.iloc[0])
+            self.pattern_emergence_times.append(0)
 
-#         except sp.CalledProcessError as e:
-#             print(f"Failed: {e}")
-#         #     if clean:
-#         #         os_remove(data_file)
-#         #     # raise Binary_crashed()
-#         # if clean:
-#         #     os_remove(data_file)
+        cluster_llks = np.zeros(shape=(self.n_clusters, num_predictions), dtype=np.float32)
 
-#         df_toreturn.index = data.index
-#         return df_toreturn
+        for i, row in data.iterrows():
+            upper_bounds = self.PFSA_llk_means + (self.PFSA_llk_stds * self.anomaly_sensitivity)
+            for j in range(self.n_clusters):
+                curr_alphabet_size = len(self.cluster_PFSA_info[j]['%SYM_FRQ'])
+                if row['alphabet_size'] > curr_alphabet_size:
+                    cluster_llks[j][i] = float('inf')
+                    continue
+                cluster_llks[j][i] = Llk(data=[row.drop(index=['alphabet_size'])], pfsafile=self.cluster_PFSA_files[j]).run()[0]
 
-#     def predict(self, X=None, *, clean=True):
-#         """ Predict whether a time series sequence is anomalous
+            # consider to be anomaly if all llks above specified upper bound (X standard deviations above the mean)
+            # in this case, we create a new model based on the anomalous pattern
+            if np.all(cluster_llks.T[i] > upper_bounds):
+                self.pattern_emergence_times.append(i)
+                # print(i)
+                self.__add_to_PFSA_library(row)
+                # print(i, self.n_clusters)
+                cluster_llks = np.append(cluster_llks, np.zeros((1, num_predictions))).reshape((self.n_clusters, num_predictions))
+                cluster_llks[self.n_clusters - 1][i] = Llk(data=[row.drop(index=['alphabet_size'])], pfsafile=self.cluster_PFSA_files[self.n_clusters-1]).run()[0]
 
-#         Args:
-#             X (pd.DataFrame or pd.Series, optional): time series data to find anomalies, if None then
-#                 predicts on original data (Default = None)
-#             clean (bool, optional): whether to remove temp files (Default = True)
+        upper_bounds = self.PFSA_llk_means + (self.PFSA_llk_stds * self.anomaly_sensitivity)
 
-#         Returns:
-#             bool or list[bool]: True if time series is an anomaly, False otherwise output shape depends on input
-#         """
+        self.cluster_llks = cluster_llks
+        # self.closest_match = np.argmin(cluster_llks, axis=0)
+        # self.predicted_active_pfsas = {}
+        # for i, pfsa in enumerate(self.closest_match):
+        #     if pfsa in self.predicted_active_pfsas:
+        #         self.predicted_active_pfsas[pfsa].append(i)
+        #     else:
+        #         self.predicted_active_pfsas[pfsa] = [i]
 
-#         if not self.fitted:
-#             raise ValueError("Model has not been fit yet")
+        self.quantized_data = pd.concat([self.quantized_data, data.drop(columns=['alphabet_size'])], axis=0).reset_index(drop=True)
 
-#         data = None
-#         num_predictions = 0
-#         # commonly want to find anomalies in original data
-#         if X is None:
-#             # occurs when model is loaded from file
-#             if self.quantized_data is None:
-#                 raise ValueError("Original data not found. Pass data to predict().")
-#             data = self.quantized_data.drop(columns=["cluster"], axis=1)
-#             num_predictions = self.quantized_data.shape[0]
-#         else:
-#             data = self.__quantize(X)
-#             num_predictions = 1 if type(X) is pd.Series else data.shape[0]
+        return self
 
-#         cluster_llks = np.empty(shape=(self.n_clusters, num_predictions), dtype=np.float32)
-#         for i in range(self.n_clusters):
-#             # print(data)
-#             # print(i, self.cluster_PFSA_files[i])
-#             # print(Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run())
-#             # print("LLK RAN")
-#             # cluster_llks[i] = np.asarray(Llk(data=data, pfsafile=self.cluster_PFSA_files[i]).run(), dtype=np.float32)
-#             cluster_llks[i] = np.asarray(llk(data, self.cluster_PFSA_files[i]), dtype=np.float32)
 
-#         # consider to be anomaly if all llks above specified upper bound (X standard deviations above the mean)
-#         upper_bounds = self.PFSA_llk_means + (self.PFSA_llk_stds * self.anomaly_sensitivity)
-#         predictions = np.all(cluster_llks.T > upper_bounds, axis=1)
+    def __add_to_PFSA_library(self, data):
+        """ Infer PFSA from data using genESeSS and update the PFSA library """
 
-#         self.cluster_llks = cluster_llks
-#         self.closest_match = np.argmin(cluster_llks, axis=0)
+        cluster_PFSA_file = RANDOM_NAME(path=self.temp_dir)
+        PFSA_png = RANDOM_NAME(path=self.temp_dir)
+        
+        alg = GenESeSS(
+            data=[data.drop(index=['alphabet_size'])],
+            outfile=cluster_PFSA_file,
+            data_type='symbolic',
+            data_dir='row',
+            force=True,
+            eps=self.eps,
+            dot=PFSA_png,
+            configfile='./patternly/config.cfg'
+        )
 
-#         if len(predictions) == 1:
-#             predictions = predictions[0]
+        # TODO: add error handling for when genESeSS fails to find a PFSA
+        PFSA_found = alg.run()
+        if PFSA_found:
+            prop_vals = [
+                alg.inference_error,
+                alg.epsilon_used,
+                alg.synchronizing_string_found,
+                alg.symbol_frequency,
+                alg.probability_morph_matrix,
+                alg.connectivity_matrix,
+            ]
+            self.cluster_PFSA_info.append(dict(zip(self.PFSA_prop_titles, prop_vals)))
+        else:
+            raise ValueError(f"Unable to find new PFSA for anomaly.")
+
+        self.cluster_PFSA_files.append(cluster_PFSA_file)
+        self.cluster_PFSA_pngs.append(PFSA_png)
+        self.n_clusters += 1
+
+        seqs = pd.DataFrame(Prun(pfsafile=cluster_PFSA_file, data_len=self.window_size, num_repeats=100).run())
+        new_data = pd.concat([pd.DataFrame(data.drop(index=['alphabet_size'])).T, seqs], axis=0)
+        llks = np.asarray(Llk(data=new_data, pfsafile=self.cluster_PFSA_files[self.n_clusters-1]).run())
+        self.PFSA_llk_means = np.mean(llks)
+        self.PFSA_llk_stds = np.std(llks, ddof=1)
 
-#         return predictions
